from __future__ import annotations

import os
import pathlib
from typing import Callable

from dynaconf import Dynaconf, Validator

DEFAULT_SETTINGS_PATH: pathlib.Path = (
    pathlib.Path(__file__).resolve().parent / "settings.toml"
)
USER_SETTINGS_PATH: pathlib.Path = (
    pathlib.Path(
        os.getenv("LIXINGER_SETTINGS_PATH", "~/.config/lixinger/settings.toml")
    )
    .expanduser()
    .resolve()
)


settings_cast_map = {}


def cast(v: str) -> Callable:
    """Decorator to register cast function for a setting."""

    def decorator(func) -> Callable:
        settings_cast_map[v] = func
        return func

    return decorator


class TypedDynaconf(Dynaconf):
    url: str
    token: str


def get_validators() -> list[Validator]:
    """Get validators for settings.

    Cast settings to their types.
    """
    settings_types = vars(TypedDynaconf)["__annotations__"]
    return [
        Validator(key, cast=cast(settings_cast_map.get(key, type_)))
        for key, type_ in settings_types.items()
    ]


def load_settings(
    path: pathlib.Path = DEFAULT_SETTINGS_PATH,
    user_settings_path: pathlib.Path = USER_SETTINGS_PATH,
) -> TypedDynaconf:
    """Load settings from default settings and user settings file."""
    settings_files = (
        [path] if not user_settings_path.exists() else [path, user_settings_path]
    )
    default_settings = TypedDynaconf(
        settings_files=settings_files,
        merge_enabled=True,
        environments=True,
        validators=get_validators(),
        validate=True,
        envvar_prefix="LIXINGER",
    )
    return default_settings


settings = load_settings()



from __future__ import annotations

from typing import Literal

import pandera as pa

from lixinger import client
from lixinger.config import settings
from lixinger.utils import api, get_response_df


class Output(pa.DataFrameModel):
    date: pa.typing.Series[pa.typing.DateTime]
    stock_code: pa.typing.Series[str]


@api
def get_fundamental_statistics(
    stock_codes: list[str],
    metrics_list: list[str],
    granularity: Literal["y10", "y5", "y3"],
) -> pa.typing.DataFrame[Output]:
    """获取基本面统计数据.

    参考文档: https://www.lixinger.com/open/api/doc?api-key=cn/company/fundamental-statistics
    """
    payload = {
        "token": settings.token,
        "stockCodes": stock_codes,
        "metricsList": metrics_list,
        "granularity": granularity,
    }

    response = client.post(
        f"{settings.base_url}/cn/company/fundamental-statistics",
        json=payload,
    )
    return get_response_df(response, Output)



import requests
from requests import Response, request
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_fixed,
)


@retry(
    stop=(stop_after_delay(60) | stop_after_attempt(3)),
    wait=wait_fixed(3),
    retry=retry_if_exception_type(
        (requests.exceptions.Timeout, requests.exceptions.ProxyError)
    ),
)
def post(url: str, data=None, json=None, **kwargs: any) -> Response:
    return request("post", url, data=data, json=json, **kwargs)



from lixinger.api.cn.index.drawdown import get_index_drawdown


def test_get_index_drawdown():
    get_index_drawdown(
        start_date="2023-04-20",
        stock_code="000300",
        granularity="m",
    )



from __future__ import annotations

import pandas as pd
import pandera as pa

from lixinger import client
from lixinger.config import settings
from lixinger.utils import api, get_response_df, set_column_snake_case


class Output(pa.DataFrameModel):
    index_code: pa.typing.Series[str]
    stock_code: pa.typing.Series[str]
    aera_code: pa.typing.Series[str]
    market: pa.typing.Series[str]


@api
def get_index_constituents(
    date: str,
    stock_codes: list[str] = None,
) -> pa.typing.DataFrame[Output]:
    """获取样本信息.

    参考文档: https://www.lixinger.com/open/api/doc?api-key=cn/index/constituents
    """
    payload = {
        "token": settings.token,
        "date": date,
    }
    if stock_codes is not None:
        payload["stockCodes"] = stock_codes
    response = client.post(
        f"{settings.base_url}/cn/index/constituents",
        json=payload,
    )
    df = get_response_df(response, Output)
    del df["index_code"]
    del df["market"]
    del df["aera_code"]
    df = df.explode("constituents")
    df = pd.concat(
        [df.drop(["constituents"], axis=1), df["constituents"].apply(pd.Series)], axis=1
    )
    df.rename(columns={"stock_code": "index_code"}, inplace=True)
    df = set_column_snake_case(df)
    return df



from lixinger.api.cn.index.fundamental import get_index_fundamental


def test_get_index_fundamental():
    get_index_fundamental(
        start_date="2023-04-20",
        end_date="2023-04-30",
        stock_codes=["000300"],
        metrics_list=["pe_ttm.mcw"],
    )



from lixinger.api.cn.company.fundamental_non_financial import (
    get_fundamental_non_financial,
)


def test_fundamental_non_financial() -> None:
    get_fundamental_non_financial(
        stock_codes=["600519"], metrics_list=["pe_ttm"], start_date="2010-01-01"
    )



from lixinger.api.cn.fund.exchange_traded_close_price import (
    get_exchange_traded_close_price,
)


def test_get_exchange_traded_close_price() -> None:
    get_exchange_traded_close_price(start_date="2021-01-01", stock_code="161725")



from lixinger.api.cn.company.fundamental_statistics import (
    get_fundamental_statistics,
)


def test_fundamental_statistics() -> None:
    get_fundamental_statistics(
        stock_codes=["300750", "600519"], granularity="y3", metrics_list=["pe_ttm"]
    )



from __future__ import annotations

import inspect
import json
import re
from collections import namedtuple
from functools import lru_cache, wraps
from typing import Callable, Type

import numpy as np
import pandas as pd
import pandera as pa
import six
from pydantic import validate_arguments
from requests import Response

from lixinger.config import settings


def camel_case_to_snake_case(name: str) -> str:
    """Convert camel case to snake case."""
    pattern = re.compile(r"(?<!^)(?=[A-Z])")
    name = pattern.sub("_", name).lower()
    return name


def set_column_snake_case(df: pd.DataFrame) -> pd.DataFrame:
    """Set dataframe column names to snake case."""
    df.columns = [camel_case_to_snake_case(col) for col in df.columns]
    return df


def convert_column_list_to_str(column: list[str]) -> str:
    """Convert column list to string."""
    try:
        return ",".join(map(str, column))
    except TypeError:
        return np.NaN


def get_response_data(response: Response) -> list[dict[str, any]]:
    """Get response data from response."""
    resp_json = response.json()
    if resp_json.get("code") != 1:
        raise ValueError(f"[{resp_json.get('code')}]{resp_json.get('error')}")
    return resp_json["data"]


def get_response_df(
    response: Response, output: Type[pa.DataFrameModel]
) -> pd.DataFrame:
    """Get response dataframe from response."""
    data = get_response_data(response)
    dtypes = {k: str(v) for k, v in output.to_schema().dtypes.items()}
    if not data:
        df = pd.DataFrame(columns=[*dtypes]).astype(dtypes)
    else:
        df = pd.json_normalize(data)
        df = set_column_snake_case(df)
        missing_columns = set(dtypes.keys()) - set(df.columns)
        if missing_columns:
            df[list(missing_columns)] = np.NaN
        df = df.astype(dtypes, errors="ignore")
        for col, dtype in dtypes.items():
            if dtype == "datetime64[ns]":
                df[col] = pd.to_datetime(df[col]).dt.tz_localize(None)
    return df


def adjust_request_date_range(func: Callable) -> Callable:
    """Adjust request date range."""

    @wraps(func)
    def wrapper(*args: any, **kwargs: any) -> Callable:
        start_date_str = kwargs.get("start_date")
        end_date_str = kwargs.get("end_date")
        limit = kwargs.get("limit")

        if start_date_str is None:
            raise ValueError("start_date is required")
        start_date = pd.Timestamp(start_date_str)

        if end_date_str is None:
            end_date = pd.Timestamp("today")
        else:
            end_date = pd.Timestamp(end_date_str)

        if start_date > end_date:
            raise ValueError("start_date should be less than end_date")

        dfs = pd.DataFrame()
        while start_date < end_date:
            new_end_date = start_date + pd.DateOffset(years=10)
            query_params = {
                **kwargs,
                "start_date": start_date.strftime("%Y-%m-%d"),
                "end_date": new_end_date.strftime("%Y-%m-%d"),
            }
            df = func(*args, **query_params)
            dfs = pd.concat([dfs, df])
            start_date = new_end_date + pd.Timedelta(days=1)

            if limit is not None:
                limit -= len(df)
                if limit <= 0:
                    break
                kwargs["limit"] = limit
        if "date" in dfs.columns:
            dfs.sort_values(by="date", inplace=True)
        return dfs

    return wrapper


Serialized = namedtuple("Serialized", "json")


def hashable_cache(func: Callable | None = None, *, maxsize=16) -> Callable:
    """Hashable cache."""

    def hashable_cache_internal(_func: Callable) -> Callable:
        cache = lru_cache(maxsize)

        def deserialize(value) -> any:
            if isinstance(value, Serialized):
                return json.loads(value.json)
            else:
                return value

        def func_with_serialized_params(*args: any, **kwargs: any) -> Callable:
            _args = tuple([deserialize(arg) for arg in args])
            _kwargs = {k: deserialize(v) for k, v in six.viewitems(kwargs)}
            return _func(*_args, **_kwargs)

        cached_func = cache(func_with_serialized_params)

        @wraps(_func)
        def hashable_cached_func(*args: any, **kwargs: any) -> Callable:
            _args = tuple(
                [
                    Serialized(json.dumps(arg, sort_keys=True))
                    if type(arg) in (list, dict)
                    else arg
                    for arg in args
                ]
            )
            _kwargs = {
                k: Serialized(json.dumps(v, sort_keys=True))
                if type(v) in (list, dict)
                else v
                for k, v in kwargs.items()
            }
            return cached_func(*_args, **_kwargs)

        hashable_cached_func.cache_info = cached_func.cache_info
        hashable_cached_func.cache_clear = cached_func.cache_clear
        return hashable_cached_func

    if func is not None:
        return hashable_cache_internal(func)
    return hashable_cache_internal


def api(func: Callable | None = None, *, maxsize=16) -> Callable:
    """API decorator."""

    def wrapper(_func: Callable) -> Callable:
        @hashable_cache(maxsize=maxsize)
        @validate_arguments
        @wraps(_func)
        def _api(*args: any, **kwargs: any) -> Callable:
            parameters = inspect.signature(_func).parameters
            if "start_date" in parameters and "end_date" in parameters:
                return adjust_request_date_range(_func)(*args, **kwargs)
            return _func(*args, **kwargs)

        return _api

    if func is not None:
        return wrapper(func)
    return wrapper


def set_token(token: str) -> None:
    settings.token = token



from __future__ import annotations

import pandera as pa

from lixinger import client
from lixinger.config import settings
from lixinger.utils import api, get_response_df


class Output(pa.DataFrameModel):
    date: pa.typing.Series[pa.typing.DateTime]
    stock_code: pa.typing.Series[str]


@api
def get_fundamental_non_financial(
    stock_codes: list[str],
    metrics_list: list[str],
    date: str | None = None,
    start_date: str | None = None,
    end_date: str | None = None,
    adjust_forward_date: str | None = None,
    adjust_backward_date: str | None = None,
    limit: int | None = None,
) -> pa.typing.DataFrame[Output]:
    """获取基本面数据，如PE、PB等.

    参考文档: https://www.lixinger.com/open/api/doc?api-key=cn/company/fundamental/non_financial
    """
    payload = {
        "token": settings.token,
        "stockCodes": stock_codes,
        "metricsList": metrics_list,
    }
    if date is not None:
        payload["date"] = date
    if start_date is not None:
        payload["startDate"] = start_date
    if end_date is not None:
        payload["endDate"] = end_date
    if adjust_forward_date is not None:
        payload["adjustForwardDate"] = adjust_forward_date
    if adjust_backward_date is not None:
        payload["adjustBackwardDate"] = adjust_backward_date
    if limit is not None:
        payload["limit"] = limit

    response = client.post(
        f"{settings.base_url}/cn/company/fundamental/non_financial",
        json=payload,
    )
    return get_response_df(response, Output)
